# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
import pytest
from io import BytesIO
from ezdxf.recover import (
    bytes_loader,
    detect_encoding,
    synced_bytes_loader,
    _detect_dxf_version,
    _search_int,
    _search_float,
    byte_tag_compiler,
)
from ezdxf.lldxf import const
from ezdxf.audit import AuditError

HEADER = """  0
SECTION
  2
HEADER
  9
$ACADVER
  1
AC1027
  9
$ACADMAINTVER
 70
105
  9
$DWGCODEPAGE
  3
ANSI_1252
  9
TEST
  1
ÄÖÜ
"""

MALFORMED_GROUP_CODES = b"""  0x
SECTION
  +2x
HEADER
  \t9\t
$ACADVER
  1.0
AC1027
"""


class TestBytesLoader:
    @pytest.fixture(params=[bytes_loader, synced_bytes_loader])
    def loader(self, request):
        return request.param

    @pytest.fixture
    def data(self):
        return BytesIO(HEADER.encode("latin1"))

    def test_bytes_loader(self, data, loader):
        result = list(loader(data))
        assert len(result) == 10
        assert result[-1] == (1, b"\xc4\xd6\xdc")

    def test_encoding_detector(self, data, loader):
        assert detect_encoding(loader(data)) == "utf8"

    def test_windows_line_endings_CR_LF(self, loader):
        tags = list(loader(BytesIO(b"0\r\n1\r\n0\r\n2\r\n")))
        assert tags == [
            (0, b"1"),
            (0, b"2"),
        ], "Windows line endings CR/LF must be supported."

    def test_linux_and_macos_x_line_endings_LF(self, loader):
        tags = list(bytes_loader(BytesIO(b"0\n1\n0\n2\n")))
        assert tags == [
            (0, b"1"),
            (0, b"2"),
        ], "Linux and MacOS X line endings LF must be supported."

    def test_line_endings_only_CR(self):
        assert (
            list(synced_bytes_loader(BytesIO(b"0\r1\r0\r2\r"))) == []
        ), "MacOS prior to MacOS X line endings CR are not supported."

    def test_malformed_group_codes(self, loader):
        data = BytesIO(MALFORMED_GROUP_CODES)
        tags = list(loader(data))
        assert tags[0] == (0, b"SECTION")
        assert tags[1] == (2, b"HEADER")
        assert tags[2] == (9, b"$ACADVER")
        assert tags[3] == (1, b"AC1027")


MALFORMED_VALUE_TAGS = b"""  70 # int value
  42xyz
40 # float value
  47.11xyz
10 # 3D vertex
 1.0x
20
 1.0y
30
 1.0z
11 # 2D vertex
 2.0x
21
 2.0y
 0
DUMMY
"""


class TestByteTagCompiler:
    def test_malformed_value_tags(self):
        loader = bytes_loader(BytesIO(MALFORMED_VALUE_TAGS))
        msg = []
        tags = list(byte_tag_compiler(loader, messages=msg))
        assert tags[0] == (70, 42)
        assert tags[1] == (40, 47.11)
        assert tags[2] == (10, (1, 1, 1))
        assert tags[3] == (11, (2, 2))
        assert len(msg) == 7

    def test_malformed_integer_tags(self):
        loader = bytes_loader(BytesIO(b"70\n123x\n"))
        msg = []
        tags = list(byte_tag_compiler(loader, messages=msg))
        assert tags[0] == (70, 123)
        assert msg[0][0] == AuditError.INVALID_INTEGER_VALUE

    def test_underlines_in_numbers_are_handled_by_python(self):
        loader = bytes_loader(BytesIO(b"50\n123_000.9\n70\n456_000\n"))
        msg = []
        tags = list(byte_tag_compiler(loader, messages=msg))
        assert tags[0] == (50, 123000.9)
        assert tags[1] == (70, 456000)
        assert len(msg) == 0

    @pytest.mark.parametrize("b", [
        b"50\n1.000 e20\n",
        b"50\n1.000 E20\n",
        b"50\n1.000\te20\n",
        b"50\n1.000\tE20\n",
    ])
    def test_ignore_whitespace_in_floats(self, b):
        loader = bytes_loader(BytesIO(b))
        msg = []
        tags = list(byte_tag_compiler(loader, messages=msg))
        assert tags[0] == (50, 1e20)
        assert msg[0][0] == AuditError.INVALID_FLOATING_POINT_VALUE


OUT_OF_SYNC_TAGS = """
  0
SECTION
  2
HEADER

  9
$ACADVER
  1
AC1027
  9
$ACADMAINTVER
 70
105
  9
$DWGCODEPAGE

  3
ANSI_1252
  9
TEST
  1
test
"""


def test_out_of_sync_tags():
    result = list(synced_bytes_loader(BytesIO(OUT_OF_SYNC_TAGS.encode())))
    assert len(result) == 10


class TestDetectDXFVersion:
    def test_missing_dxf_version_is_r12(self):
        assert _detect_dxf_version([]) == "AC1009"

    def test_detect_r12(self):
        assert _detect_dxf_version([(9, "$ACADVER"), (3, "AC1009")]) == "AC1009"

    @pytest.mark.parametrize("ver", list(const.acad_release.keys()))
    def test_detect_all_supported_versions(self, ver):
        assert _detect_dxf_version([(9, "$ACADVER"), (3, ver)]) == ver

    def test_detect_any_well_formed_version(self):
        assert _detect_dxf_version([(9, "$ACADVER"), (3, "AC9999")]) == "AC9999"

    @pytest.mark.parametrize(
        "ver", ["AC10x9", "XC1009", "0000", "AC10240", 0, None]
    )
    def test_malformed_version_is_r12(self, ver):
        assert _detect_dxf_version([(9, "$ACADVER"), (3, ver)]) == "AC1009"

    @pytest.mark.parametrize(
        "ver",
        [
            " AC1015",
            "AC1015 ",
            "  AC1015  ",
        ],
    )
    def test_strip_whitespace_from_version_string(self, ver):
        assert _detect_dxf_version([(9, "$ACADVER"), (3, ver)]) == "AC1015"


class TestSearchIntConverter:
    @pytest.mark.parametrize("s", ["0", "00", "+0", "-0"])
    def test_valid_zero_int(self, s):
        assert _search_int(s) == 0

    @pytest.mark.parametrize("s", ["+1", "1", "01"])
    def test_valid_int(self, s):
        assert _search_int(s) == 1

    @pytest.mark.parametrize("s", [" 1", "\t1", "\r1", " \r1"])
    def test_ignore_whitespace_in_front_of_int(self, s):
        assert _search_int(s) == 1

    @pytest.mark.parametrize("s", ["1 ", "1\t", "1\r", "1 \r"])
    def test_ignore_whitespace_behind_int(self, s):
        assert _search_int(s) == 1

    @pytest.mark.parametrize(
        "s",
        [
            "1.",
            "1,",
            "1a",
            "1-",
            "1+",
            "1$",
            "1 1",
            "+1 x fd fgg",
            "\t\r+1_ ",
        ],
    )
    def test_ignore_invalid_chars_behind_int(self, s):
        assert _search_int(s) == 1

    @pytest.mark.parametrize(
        "b",
        [
            b"1.",
            b"1,",
            b"1a",
            b"1-",
            b"\t\r+1_ ",
        ],
    )
    def test_if_works_with_bytes(self, b):
        assert _search_int(b) == 1


class TestSearchFloatConverter:
    @pytest.mark.parametrize("s", ["0", "0.0", "+0.", "-0."])
    def test_valid_zero(self, s):
        assert _search_float(s) == 0.0

    @pytest.mark.parametrize(
        "s",
        ["1", "+1", "1.", "1.0", "1e0", "1.e0", "1.0e0", "1E0", "1e+0", "1e-0"],
    )
    def test_valid_positive_floats(self, s):
        assert _search_float(s) == 1.0

    @pytest.mark.parametrize(
        "s",
        [
            "-1",
            "-1.",
            "-1.0",
            "-1e0",
            "-1.e0",
            "-1.0e0",
        ],
    )
    def test_valid_negative_floats(self, s):
        assert _search_float(s) == -1.0

    def test_negative_exponent(self):
        assert _search_float("1e-1") == 0.1
        assert _search_float("-1e-1") == -0.1

    @pytest.mark.parametrize("s", [" 1.0", "\t1.0", "\r1.0", " \r1.0"])
    def test_ignore_whitespace_in_front_of_float(self, s):
        assert _search_float(s) == 1.0

    @pytest.mark.parametrize("s", ["1.0 ", "1.\t", "1.\r", "1. \r", "1e0 \t"])
    def test_ignore_whitespace_behind_float(self, s):
        assert _search_float(s) == 1.0

    @pytest.mark.parametrize(
        "s",
        [
            "1,",
            "1a",
            "1-",
            "1+",
            "1$",
            "1 1",
            "+1 x fd fgg",
            "\t\r+1_ ",
            "1.,",
            "1.a",
            "1.-",
            "1.+",
            "1.$",
            "1. 1",
            "+1. x fd fgg",
            "\t\r+1._ ",
            "1e0,",
            "1e0a",
            "1e0-",
            "1e0+",
            "1e0$",
            "1e0 1",
            "+1e0 x fd fgg",
            "\t\r+1e0_ ",
            "  1e+0xx",
            "  1e-0xx",
            "  1E+0xx",
            "  1E-0xx",
        ],
    )
    def test_ignore_invalid_chars_behind_float(self, s):
        assert _search_float(s) == 1.0

    @pytest.mark.parametrize(
        "b",
        [
            b"1,",
            b"1a",
            b"1-",
            b"1+",
            b"1$",
            b"1 1",
            b"+1 x fd fgg",
            b"\t\r+1_ ",
        ],
    )
    def test_if_works_bytes(self, b):
        assert _search_float(b) == 1.0



# Source: https://github.com/mapbox/earcut
# License: ISC License (MIT compatible)
#
# Copyright (c) 2016, Mapbox
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.
#
# The Algorithm
# -------------
# The library implements a modified ear slicing algorithm, optimized by z-order
# curve hashing and extended to handle holes, twisted polygons, degeneracies and
# self-intersections in a way that doesn't guarantee correctness of triangulation,
# but attempts to always produce acceptable results for practical data.
#
# Translation to Python:
# Copyright (c) 2022, Manfred Moitzi
# License: MIT License
#
# Notes
# -----
# Exterior path (outer path) vertices are stored in counter-clockwise order
# Hole vertices are stored in clockwise order
# Vertex order will be maintained by the algorithm automatically.
# Boundary behavior for holes:
#   - holes outside the exterior path are ignored
#   - invalid result for holes partially extending beyond the exterior path
#   - invalid result for overlapping holes
#   - invalid result for holes in holes
# Very stable in all circumstances - DOES NOT CRASH!
#
# Steiner Point
# -------------
# https://en.wikipedia.org/wiki/Steiner_point_(computational_geometry)
# A Steiner point is a point that is not part of the input to a geometric
# optimization problem but is added during the solution of the problem, to
# create a better solution than would be possible from the original points
# alone.
# A Steiner point is defined as a hole with a single point!
#
from __future__ import annotations
from typing import Sequence, Optional, Protocol, TypeVar

import math


class Point(Protocol):
    x: float
    y: float


class Node:
    def __init__(self, i: int, point: Point) -> None:
        self.i: int = i

        # store source point for output
        self.point = point

        # vertex coordinates
        self.x: float = point.x
        self.y: float = point.y

        # previous and next vertex nodes in a polygon ring
        self.prev: Node = None  # type: ignore
        self.next: Node = None  # type: ignore

        # z-order curve value
        self.z: int = 0

        # previous and next nodes in z-order
        self.prev_z: Node = None  # type: ignore
        self.next_z: Node = None  # type: ignore

        # indicates whether this is a steiner point
        self.steiner: bool = False

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y


T = TypeVar("T", bound=Point)


def earcut(exterior: list[T], holes: list[list[T]]) -> list[Sequence[T]]:
    """Implements a modified ear slicing algorithm, optimized by z-order
    curve hashing and extended to handle holes, twisted polygons, degeneracies
    and self-intersections in a way that doesn't guarantee correctness of
    triangulation, but attempts to always produce acceptable results for
    practical data.

    Source: https://github.com/mapbox/earcut

    Args:
        exterior: outer path as list of points as objects which provide a
            `x`- and a `y`-attribute
        holes: list of holes, each hole is list of points, a hole with
            a single points is a Steiner point

    Returns:
        Returns a list of triangles, each triangle is a tuple of three points,
        the output points are the same objects as the input points.

    """
    # exterior points in counter-clockwise order
    outer_node: Node = linked_list(exterior, 0, ccw=True)
    triangles: list[Sequence[T]] = []

    if outer_node is None or outer_node.next is outer_node.prev:
        return triangles

    if len(holes) > 0:
        outer_node = eliminate_holes(holes, len(exterior), outer_node)

    min_x: float = 0.0
    min_y: float = 0.0
    inv_size: float = 0.0

    # if the shape is not too simple, we'll use z-order curve hash later
    # calculate polygon bbox
    if len(exterior) > 80:
        min_x = max_x = exterior[0].x
        min_y = max_y = exterior[0].y
        for point in exterior:
            x = point.x
            y = point.y
            min_x = min(min_x, x)
            min_y = min(min_y, y)
            max_x = max(max_x, x)
            max_y = max(max_y, y)

        # min_x, min_y and inv_size are later used to transform coords into
        # integers for z-order calculation
        inv_size = max(max_x - min_x, max_y - min_y)
        inv_size = 32767 / inv_size if inv_size != 0 else 0

    earcut_linked(outer_node, triangles, min_x, min_y, inv_size, 0)  # type: ignore
    return triangles


def linked_list(points: Sequence[Point], start: int, ccw: bool) -> Node:
    """Create a circular doubly linked list from polygon points in the specified
    winding order
    """
    last: Node = None  # type: ignore
    if ccw is (signed_area(points) < 0):
        for point in points:
            last = insert_node(start, point, last)
            start += 1
    else:
        end = start + len(points)
        for point in reversed(points):
            last = insert_node(end, point, last)
            end -= 1

    # open polygon: where the 1st vertex is not coincident with the last vertex
    if last and last == last.next:  # true equals
        remove_node(last)
        last = last.next
    return last


def signed_area(points: Sequence[Point]) -> float:
    s: float = 0.0
    if not len(points):
        return s
    prev = points[-1]
    for point in points:
        s += (point.x - prev.x) * (point.y + prev.y)
        prev = point
    # s < 0 is counter-clockwise
    # s > 0 is clockwise
    return s


def area(p: Node, q: Node, r: Node) -> float:
    """Returns signed area of a triangle"""
    return (q.y - p.y) * (r.x - q.x) - (q.x - p.x) * (r.y - q.y)


def is_valid_diagonal(a: Node, b: Node):
    """Check if a diagonal between two polygon nodes is valid (lies in polygon
    interior)
    """
    return (
        a.next.i != b.i
        and a.prev.i != b.i
        and not intersects_polygon(a, b)  # doesn't intersect other edges
        and (
            locally_inside(a, b)
            and locally_inside(b, a)
            and middle_inside(a, b)
            and (
                area(a.prev, a, b.prev) or area(a, b.prev, b)
            )  # does not create opposite-facing sectors
            or a == b  # true equals
            and area(a.prev, a, a.next) > 0
            and area(b.prev, b, b.next) > 0
        )  # special zero-length case
    )


def intersects_polygon(a: Node, b: Node) -> bool:
    """Check if a polygon diagonal intersects any polygon segments"""
    p = a
    while True:
        if (
            p.i != a.i
            and p.next.i != a.i
            and p.i != b.i
            and p.next.i != b.i
            and intersects(p, p.next, a, b)
        ):
            return True
        p = p.next
        if p is a:
            break
    return False


def sign(num: float) -> int:
    if num < 0.0:
        return -1
    if num > 0.0:
        return 1
    return 0


def on_segment(p: Node, q: Node, r: Node) -> bool:
    return max(p.x, r.x) >= q.x >= min(p.x, r.x) and max(p.y, r.y) >= q.y >= min(
        p.y, r.y
    )


def intersects(p1: Node, q1: Node, p2: Node, q2: Node) -> bool:
    """check if two segments intersect"""
    o1 = sign(area(p1, q1, p2))
    o2 = sign(area(p1, q1, q2))
    o3 = sign(area(p2, q2, p1))
    o4 = sign(area(p2, q2, q1))

    if o1 != o2 and o3 != o4:
        return True  # general case

    if o1 == 0 and on_segment(p1, p2, q1):
        return True  # p1, q1 and p2 are collinear and p2 lies on p1q1
    if o2 == 0 and on_segment(p1, q2, q1):
        return True  # p1, q1 and q2 are collinear and q2 lies on p1q1
    if o3 == 0 and on_segment(p2, p1, q2):
        return True  # p2, q2 and p1 are collinear and p1 lies on p2q2
    if o4 == 0 and on_segment(p2, q1, q2):
        return True  # p2, q2 and q1 are collinear and q1 lies on p2q2
    return False


def insert_node(i: int, point: Point, last: Node) -> Node:
    """create a node and optionally link it with previous one (in a circular
    doubly linked list)
    """
    p = Node(i, point)

    if last is None:
        p.prev = p
        p.next = p
    else:
        p.next = last.next
        p.prev = last
        last.next.prev = p
        last.next = p
    return p


def remove_node(p: Node) -> None:
    p.next.prev = p.prev
    p.prev.next = p.next

    if p.prev_z:
        p.prev_z.next_z = p.next_z
    if p.next_z:
        p.next_z.prev_z = p.prev_z


def eliminate_holes(
    holes: Sequence[Sequence[Point]], start: int, outer_node: Node
) -> Node:
    """link every hole into the outer loop, producing a single-ring polygon
    without holes
    """
    queue: list[Node] = []
    for hole in holes:
        if len(hole) < 1:  # skip empty holes
            continue
        # hole vertices in clockwise order
        _list = linked_list(hole, start, ccw=False)
        if _list is _list.next:
            _list.steiner = True
        start += len(hole)
        queue.append(get_leftmost(_list))
    queue.sort(key=lambda node: (node.x, node.y))

    # process holes from left to right
    for hole_ in queue:
        outer_node = eliminate_hole(hole_, outer_node)
    return outer_node


def eliminate_hole(hole: Node, outer_node: Node) -> Node:
    """Find a bridge between vertices that connects hole with an outer ring and
    link it
    """
    bridge = find_hole_bridge(hole, outer_node)
    if bridge is None:
        return outer_node

    bridge_reverse = split_polygon(bridge, hole)

    # filter collinear points around the cuts
    filter_points(bridge_reverse, bridge_reverse.next)
    return filter_points(bridge, bridge.next)


def filter_points(start: Node, end: Optional[Node] = None) -> Node:
    """eliminate colinear or duplicate points"""
    if start is None:
        return start
    if end is None:
        end = start

    p = start

    while True:
        again = False
        if not p.steiner and (
            p == p.next or area(p.prev, p, p.next) == 0  # true equals
        ):
            remove_node(p)
            p = end = p.prev
            if p is p.next:
                break
            again = True
        else:
            p = p.next
        if not (again or p is not end):
            break
    return end


# main ear slicing loop which triangulates a polygon (given as a linked list)
def earcut_linked(
    ear: Node,
    triangles: list[Sequence[Point]],
    min_x: float,
    min_y: float,
    inv_size: float,
    pass_: int,
) -> None:
    if ear is None:
        return

    # interlink polygon nodes in z-order
    if not pass_ and inv_size:
        index_curve(ear, min_x, min_y, inv_size)

    stop = ear

    # iterate through ears, slicing them one by one
    while ear.prev is not ear.next:
        prev = ear.prev
        next = ear.next

        _is_ear = (
            is_ear_hashed(ear, min_x, min_y, inv_size) if inv_size else is_ear(ear)
        )
        if _is_ear:
            # cut off the triangle
            triangles.append((prev.point, ear.point, next.point))
            remove_node(ear)

            # skipping the next vertex leads to less sliver triangles
            ear = next.next
            stop = next.next
            continue

        ear = next

        # if we looped through the whole remaining polygon and can't find any more ears
        if ear is stop:
            # try filtering points and slicing again
            if not pass_:
                earcut_linked(
                    filter_points(ear),
                    triangles,
                    min_x,
                    min_y,
                    inv_size,
                    1,
                )

            # if this didn't work, try curing all small self-intersections locally
            elif pass_ == 1:
                ear = cure_local_intersections(filter_points(ear), triangles)
                earcut_linked(ear, triangles, min_x, min_y, inv_size, 2)

            # as a last resort, try splitting the remaining polygon into two
            elif pass_ == 2:
                split_ear_cut(ear, triangles, min_x, min_y, inv_size)
            break


def is_ear(ear: Node) -> bool:
    """check whether a polygon node forms a valid ear with adjacent nodes"""
    a: Node = ear.prev
    b: Node = ear
    c: Node = ear.next

    if area(a, b, c) >= 0:
        return False  # reflex, can't be an ear

    # now make sure we don't have other points inside the potential ear
    ax = a.x
    bx = b.x
    cx = c.x
    ay = a.y
    by = b.y
    cy = c.y

    # triangle bbox; min & max are calculated like this for speed
    x0 = min(ax, bx, cx)
    x1 = max(ax, bx, cx)
    y0 = min(ay, by, cy)
    y1 = max(ay, by, cy)
    p: Node = c.next

    while p is not a:
        if (
            x0 <= p.x <= x1
            and y0 <= p.y <= y1
            and point_in_triangle(ax, ay, bx, by, cx, cy, p.x, p.y)
            and area(p.prev, p, p.next) >= 0
        ):
            return False
        p = p.next

    return True


def is_ear_hashed(ear: Node, min_x: float, min_y: float, inv_size: float):
    a: Node = ear.prev
    b: Node = ear
    c: Node = ear.next

    if area(a, b, c) >= 0:
        return False  # reflex, can't be an ear

    ax = a.x
    bx = b.x
    cx = c.x
    ay = a.y
    by = b.y
    cy = c.y

    # triangle bbox; min & max are calculated like this for speed
    x0 = min(ax, bx, cx)
    x1 = max(ax, bx, cx)
    y0 = min(ay, by, cy)
    y1 = max(ay, by, cy)

    # z-order range for the current triangle bbox;
    min_z = z_order(x0, y0, min_x, min_y, inv_size)
    max_z = z_order(x1, y1, min_x, min_y, inv_size)

    p: Node = ear.prev_z
    n: Node = ear.next_z

    # look for points inside the triangle in both directions
    while p and p.z >= min_z and n and n.z <= max_z:
        if (
            x0 <= p.x <= x1
            and y0 <= p.y <= y1
            and p is not a
            and p is not c
            and point_in_triangle(ax, ay, bx, by, cx, cy, p.x, p.y)
            and area(p.prev, p, p.next) >= 0
        ):
            return False
        p = p.prev_z

        if (
            x0 <= n.x <= x1
            and y0 <= n.y <= y1
            and n is not a
            and n is not c
            and point_in_triangle(ax, ay, bx, by, cx, cy, n.x, n.y)
            and area(n.prev, n, n.next) >= 0
        ):
            return False
        n = n.next_z

    # look for remaining points in decreasing z-order
    while p and p.z >= min_z:
        if (
            x0 <= p.x <= x1
            and y0 <= p.y <= y1
            and p is not a
            and p is not c
            and point_in_triangle(ax, ay, bx, by, cx, cy, p.x, p.y)
            and area(p.prev, p, p.next) >= 0
        ):
            return False
        p = p.prev_z

    # look for remaining points in increasing z-order
    while n and n.z <= max_z:
        if (
            x0 <= n.x <= x1
            and y0 <= n.y <= y1
            and n is not a
            and n is not c
            and point_in_triangle(ax, ay, bx, by, cx, cy, n.x, n.y)
            and area(n.prev, n, n.next) >= 0
        ):
            return False
        n = n.next_z
    return True


def get_leftmost(start: Node) -> Node:
    """Find the leftmost node of a polygon ring"""
    p = start
    leftmost = start
    while True:
        if p.x < leftmost.x or (p.x == leftmost.x and p.y < leftmost.y):
            leftmost = p
        p = p.next
        if p is start:
            break
    return leftmost


def point_in_triangle(
    ax: float,
    ay: float,
    bx: float,
    by: float,
    cx: float,
    cy: float,
    px: float,
    py: float,
) -> bool:
    """Check if a point lies within a convex triangle"""
    return (
        (cx - px) * (ay - py) >= (ax - px) * (cy - py)
        and (ax - px) * (by - py) >= (bx - px) * (ay - py)
        and (bx - px) * (cy - py) >= (cx - px) * (by - py)
    )


def sector_contains_sector(m: Node, p: Node):
    """Whether sector in vertex m contains sector in vertex p in the same
    coordinates.
    """
    return area(m.prev, m, p.prev) < 0 and area(p.next, m, m.next) < 0


def index_curve(start: Node, min_x: float, min_y: float, inv_size: float):
    """Interlink polygon nodes in z-order"""
    p = start
    while True:
        if p.z == 0:
            p.z = z_order(p.x, p.y, min_x, min_y, inv_size)
        p.prev_z = p.prev
        p.next_z = p.next
        p = p.next
        if p is start:
            break

    p.prev_z.next_z = None  # type: ignore
    p.prev_z = None  # type: ignore

    sort_linked(p)


def z_order(x0: float, y0: float, min_x: float, min_y: float, inv_size: float) -> int:
    """Z-order of a point given coords and inverse of the longer side of data
    bbox.
    """
    # coords are transformed into non-negative 15-bit integer range
    x = int((x0 - min_x) * inv_size)
    y = int((y0 - min_y) * inv_size)

    x = (x | (x << 8)) & 0x00FF00FF
    x = (x | (x << 4)) & 0x0F0F0F0F
    x = (x | (x << 2)) & 0x33333333
    x = (x | (x << 1)) & 0x55555555

    y = (y | (y << 8)) & 0x00FF00FF
    y = (y | (y << 4)) & 0x0F0F0F0F
    y = (y | (y << 2)) & 0x33333333
    y = (y | (y << 1)) & 0x55555555

    return x | (y << 1)


# Simon Tatham's linked list merge sort algorithm
# http://www.chiark.greenend.org.uk/~sgtatham/algorithms/listsort.html
def sort_linked(head: Node) -> Node:
    in_size = 1
    tail: Node
    while True:
        p = head
        head = None  # type: ignore
        tail = None  # type: ignore
        num_merges = 0
        while p:
            num_merges += 1
            q = p
            p_size = 0
            for i in range(in_size):
                p_size += 1
                q = q.next_z
                if not q:
                    break
            q_size = in_size
            while p_size > 0 or (q_size > 0 and q):
                if p_size != 0 and (q_size == 0 or not q or p.z <= q.z):
                    e = p
                    p = p.next_z
                    p_size -= 1
                else:
                    e = q
                    q = q.next_z
                    q_size -= 1

                if tail:
                    tail.next_z = e
                else:
                    head = e
                e.prev_z = tail
                tail = e
            p = q
        tail.next_z = None  # type: ignore
        in_size *= 2
        if num_merges <= 1:
            break
    return head


def split_polygon(a: Node, b: Node) -> Node:
    """Link two polygon vertices with a bridge.

    If the vertices belong to the same ring, it splits polygon into two.
    If one belongs to the outer ring and another to a hole, it merges it into a
    single ring.
    """
    a2 = Node(a.i, a.point)
    b2 = Node(b.i, b.point)
    an = a.next
    bp = b.prev

    a.next = b
    b.prev = a

    a2.next = an
    an.prev = a2

    b2.next = a2
    a2.prev = b2

    bp.next = b2
    b2.prev = bp

    return b2


# go through all polygon nodes and cure small local self-intersections
def cure_local_intersections(start: Node, triangles: list[Sequence[Point]]) -> Node:
    p = start
    while True:
        a = p.prev
        b = p.next.next

        if (
            not a == b  # true equals
            and intersects(a, p, p.next, b)
            and locally_inside(a, b)
            and locally_inside(b, a)
        ):
            triangles.append((a.point, p.point, b.point))
            # remove two nodes involved
            remove_node(p)
            remove_node(p.next)
            p = start = b

        p = p.next
        if p is start:
            break
    return filter_points(p)


def split_ear_cut(
    start: Node,
    triangles: list[Sequence[Point]],
    min_x: float,
    min_y: float,
    inv_size: float,
) -> None:
    """Try splitting polygon into two and triangulate them independently"""
    # look for a valid diagonal that divides the polygon into two
    a = start
    while True:
        b = a.next.next
        while b is not a.prev:
            if a.i != b.i and is_valid_diagonal(a, b):
                # split the polygon in two by the diagonal
                c = split_polygon(a, b)

                # filter colinear points around the cuts
                a = filter_points(a, a.next)
                c = filter_points(c, c.next)

                # run earcut on each half
                earcut_linked(a, triangles, min_x, min_y, inv_size, 0)
                earcut_linked(c, triangles, min_x, min_y, inv_size, 0)
                return
            b = b.next
        a = a.next
        if a is start:
            break


# David Eberly's algorithm for finding a bridge between hole and outer polygon
def find_hole_bridge(hole: Node, outer_node: Node) -> Node:
    p = outer_node
    hx = hole.x
    hy = hole.y
    qx = -math.inf
    m: Node = None  # type: ignore
    # find a segment intersected by a ray from the hole's leftmost point to the left;
    # segment's endpoint with lesser x will be potential connection point
    while True:
        if p.y >= hy >= p.next.y != p.y:
            x = p.x + (hy - p.y) * (p.next.x - p.x) / (p.next.y - p.y)
            if hx >= x > qx:
                qx = x
                m = p if p.x < p.next.x else p.next
                if x == hx:  # ??? use math.isclose
                    # hole touches outer segment; pick leftmost endpoint
                    return m
        p = p.next
        if p is outer_node:
            break

    if m is None:
        return None

    # look for points inside the triangle of hole point, segment intersection and endpoint;
    # if there are no points found, we have a valid connection;
    # otherwise choose the point of the minimum angle with the ray as connection point
    stop = m
    mx = m.x
    my = m.y
    tan_min = math.inf
    p = m

    while True:
        if (
            hx >= p.x >= mx
            and hx != p.x
            and point_in_triangle(
                hx if hy < my else qx,
                hy,
                mx,
                my,
                qx if hy < my else hx,
                hy,
                p.x,
                p.y,
            )
        ):
            tan = abs(hy - p.y) / (hx - p.x)  # tangential

            if locally_inside(p, hole) and (
                tan < tan_min
                or (
                    tan == tan_min
                    and (p.x > m.x or (p.x == m.x and sector_contains_sector(m, p)))
                )
            ):
                m = p
                tan_min = tan

        p = p.next
        if p is stop:
            break
    return m


def locally_inside(a: Node, b: Node) -> bool:
    """Check if a polygon diagonal is locally inside the polygon"""
    return (
        area(a, b, a.next) >= 0 and area(a, a.prev, b) >= 0
        if area(a.prev, a, a.next) < 0
        else area(a, b, a.prev) < 0 or area(a, a.next, b) < 0
    )


def middle_inside(a: Node, b: Node) -> bool:
    """Check if the middle point of a polygon diagonal is inside the polygon"""
    p = a
    inside = False
    px = (a.x + b.x) / 2
    py = (a.y + b.y) / 2
    while True:
        if (
            ((p.y > py) != (p.next.y > py))
            and p.next.y != p.y
            and (px < (p.next.x - p.x) * (py - p.y) / (p.next.y - p.y) + p.x)
        ):
            inside = not inside
        p = p.next
        if p is a:
            break
    return inside



#  Copyright (c) 2020-2024, Manfred Moitzi
#  License: MIT License
import sys
from ezdxf._options import options

# Set environment variable EZDXF_DISABLE_C_EXT to '1' or 'True' to disable
# the usage of C extensions implemented by Cython.
#
# Important: If you change the EZDXF_DISABLE_C_EXT state, you have to restart
# the Python interpreter, because C extension integration is done at the
# ezdxf import!
#
# Config files:
# Section: core
# Key: disable_c_ext = 1
#
# Direct imports from the C extension modules can not be disabled,
# just the usage by the ezdxf core package.
# For an example see ezdxf.math.__init__, if you import Vec3 from ezdxf.math
# the implementation depends on DISABLE_C_EXT and the existence of the C
# extension, but if you import Vec3 from ezdxf.math.vectors, you always get
# the Python implementation.

USE_C_EXT = not options.disable_c_ext

# C-extensions are always disabled for pypy because JIT compiled Python code is
# much faster!
PYPY = hasattr(sys, 'pypy_version_info')
if PYPY:
    USE_C_EXT = False

if USE_C_EXT:
    try:
        from ezdxf.acc import vector
    except ImportError:
        USE_C_EXT = False

# set actual state of C-extension usage
options._use_c_ext = USE_C_EXT



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import NamedTuple
# A Visual Guide to the Anatomy of Typography: https://visme.co/blog/type-anatomy/
# Anatomy of a Character: https://www.fonts.com/content/learning/fontology/level-1/type-anatomy/anatomy


class FontMeasurements(NamedTuple):
    baseline: float
    cap_height: float
    x_height: float
    descender_height: float

    def scale(self, factor: float = 1.0) -> FontMeasurements:
        return FontMeasurements(
            self.baseline * factor,
            self.cap_height * factor,
            self.x_height * factor,
            self.descender_height * factor,
        )

    def shift(self, distance: float = 0.0) -> FontMeasurements:
        return FontMeasurements(
            self.baseline + distance,
            self.cap_height,
            self.x_height,
            self.descender_height,
        )

    def scale_from_baseline(self, desired_cap_height: float) -> FontMeasurements:
        factor = desired_cap_height / self.cap_height
        return FontMeasurements(
            self.baseline,
            desired_cap_height,
            self.x_height * factor,
            self.descender_height * factor,
        )

    @property
    def cap_top(self) -> float:
        return self.baseline + self.cap_height

    @property
    def x_top(self) -> float:
        return self.baseline + self.x_height

    @property
    def bottom(self) -> float:
        return self.baseline - self.descender_height

    @property
    def total_height(self) -> float:
        return self.cap_height + self.descender_height



# Copyright (c) 2020-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import cast
import math
from ezdxf.entities import factory, Point, DXFGraphic
from ezdxf.math import Vec3, UCS, NULLVEC


def virtual_entities(
    point: Point, pdsize: float = 1, pdmode: int = 0
) -> list[DXFGraphic]:
    """Yields point graphic as DXF primitives LINE and CIRCLE entities.
    The dimensionless point is rendered as zero-length line!

    Check for this condition::

        e.dxftype() == 'LINE' and e.dxf.start.isclose(e.dxf.end)

    if the rendering engine can't handle zero-length lines.


    Args:
        point: DXF POINT entity
        pdsize: point size in drawing units
        pdmode: point styling mode, see :class:`~ezdxf.entities.Point` class

    """

    def add_line_symmetrical(offset: Vec3):
        dxfattribs["start"] = ucs.to_wcs(-offset)
        dxfattribs["end"] = ucs.to_wcs(offset)
        entities.append(cast(DXFGraphic, factory.new("LINE", dxfattribs)))

    def add_line(s: Vec3, e: Vec3):
        dxfattribs["start"] = ucs.to_wcs(s)
        dxfattribs["end"] = ucs.to_wcs(e)
        entities.append(cast(DXFGraphic, factory.new("LINE", dxfattribs)))

    center = point.dxf.location
    # This is not a real OCS! Defines just the point orientation,
    # location is in WCS!
    ocs = point.ocs()
    ucs = UCS(origin=center, ux=ocs.ux, uz=ocs.uz)

    # The point angle is clockwise oriented:
    ucs = ucs.rotate_local_z(math.radians(-point.dxf.angle))

    entities: list[DXFGraphic] = []
    gfx = point.graphic_properties()

    radius = pdsize * 0.5
    has_circle = bool(pdmode & 32)
    has_square = bool(pdmode & 64)
    style = pdmode & 7

    dxfattribs = dict(gfx)
    if style == 0:  # . dimensionless point as zero-length line
        add_line_symmetrical(NULLVEC)
    # style == 1: no point symbol
    elif style == 2:  # + cross
        add_line_symmetrical(Vec3(pdsize, 0))
        add_line_symmetrical(Vec3(0, pdsize))
    elif style == 3:  # x cross
        add_line_symmetrical(Vec3(pdsize, pdsize))
        add_line_symmetrical(Vec3(pdsize, -pdsize))
    elif style == 4:  # ' tick
        add_line(NULLVEC, Vec3(0, radius))
    if has_square:
        x1 = -radius
        x2 = radius
        y1 = -radius
        y2 = radius
        add_line(Vec3(x1, y1), Vec3(x2, y1))
        add_line(Vec3(x2, y1), Vec3(x2, y2))
        add_line(Vec3(x2, y2), Vec3(x1, y2))
        add_line(Vec3(x1, y2), Vec3(x1, y1))
    if has_circle:
        dxfattribs = dict(gfx)
        if point.dxf.hasattr("extrusion"):
            dxfattribs["extrusion"] = ocs.uz
            dxfattribs["center"] = ocs.from_wcs(center)
        else:
            dxfattribs["center"] = center
        dxfattribs["radius"] = radius
        entities.append(cast(DXFGraphic, factory.new("CIRCLE", dxfattribs)))

    return entities



#  Copyright (c) 2021-2022, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Union, Iterable, TYPE_CHECKING
from pathlib import Path
from ezdxf.lldxf import loader
from ezdxf.lldxf.types import DXFTag
from ezdxf.lldxf.tagger import ascii_tags_loader, binary_tags_loader
from ezdxf.lldxf.validator import is_dxf_file, is_binary_dxf_file
from ezdxf.filemanagement import dxf_file_info

if TYPE_CHECKING:
    from ezdxf.eztypes import SectionDict


def load_section_dict(filename: Union[str, Path]) -> SectionDict:
    tagger = get_tag_loader(filename)
    return loader.load_dxf_structure(tagger)


def get_tag_loader(
    filename: Union[str, Path], errors: str = "ignore"
) -> Iterable[DXFTag]:

    filename = str(filename)
    if is_binary_dxf_file(filename):
        with open(filename, "rb") as fp:
            data = fp.read()
            return binary_tags_loader(data, errors=errors)

    if not is_dxf_file(filename):
        raise IOError(f"File '{filename}' is not a DXF file.")

    info = dxf_file_info(filename)
    with open(filename, mode="rt", encoding=info.encoding, errors=errors) as fp:
        return list(ascii_tags_loader(fp, skip_comments=True))



# Copyright (c) 2011-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Iterable,
    Iterator,
    Union,
    cast,
    Optional,
)
import logging
from pyparsing import ParseException
from ezdxf.audit import Auditor, AuditError
from ezdxf.layouts.blocklayout import BlockLayout
from ezdxf.lldxf import const, validator
from ezdxf.lldxf.const import (
    DXFBlockInUseError,
    DXFKeyError,
    DXFStructureError,
    DXFTableEntryError,
    DXFTypeError,
)
from ezdxf.entities import (
    Attrib,
    Block,
    BlockRecord,
    EndBlk,
    entity_linker,
    factory,
    is_graphic_entity,
)
from ezdxf.math import UVec, NULLVEC, Vec3
from ezdxf.render.arrows import ARROWS

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.entities import DXFEntity, DXFTagStorage
    from ezdxf.entitydb import EntityDB
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.sections.table import Table

logger = logging.getLogger("ezdxf")


def is_special_block(name: str) -> bool:
    name = name.upper()
    # Anonymous dimension, groups and table blocks do not have explicit
    # references by an INSERT entity:
    if is_anonymous_block(name):
        return True

    # Arrow blocks maybe used in DIMENSION or LEADER override without an
    # INSERT reference:
    if ARROWS.is_ezdxf_arrow(name):
        return True
    if name.startswith("_"):
        if ARROWS.is_acad_arrow(ARROWS.arrow_name(name)):
            return True

    return False


def is_anonymous_block(name: str) -> bool:
    # *U### = anonymous BLOCK, require an explicit INSERT to be in use
    # *E### = anonymous non-uniformly scaled BLOCK, requires INSERT?
    # *X### = anonymous HATCH graphic, requires INSERT?
    # *D### = anonymous DIMENSION graphic, has no explicit INSERT
    # *A### = anonymous GROUP, requires INSERT?
    # *T### = anonymous block for ACAD_TABLE, has no explicit INSERT
    return len(name) > 1 and name[0] == "*" and name[1] in "UEXDAT"


def recover_block_name(block: Block) -> str:
    name = block.dxf.get("name", "")
    if name:
        return name
    owner = block.dxf.get("owner", "")
    if not owner:
        return ""
    doc = block.doc
    # The owner of BLOCK is BLOCK_RECORD which also stores the block name
    # as group code 2; DXF attribute name is "name"
    if doc is not None and doc.entitydb is not None:
        block_record = doc.entitydb.get(owner)
        if isinstance(block_record, BlockRecord):
            return block_record.dxf.get("name", "")
    return ""


_MISSING_BLOCK_ = Block()


class BlocksSection:
    """
    Manages BLOCK definitions in a dict(), block names are case insensitive
    e.g. 'Test' == 'TEST'.

    """

    def __init__(
        self,
        doc: Optional[Drawing] = None,
        entities: Optional[list[DXFEntity]] = None,
    ):
        self.doc = doc
        if entities is not None:
            self.load(entities)
        self._reconstruct_orphaned_block_records()
        self._anonymous_block_counter = 0

    def __len__(self):
        return len(self.block_records)

    @staticmethod
    def key(entity: Union[str, BlockLayout]) -> str:
        if not isinstance(entity, str):
            entity = entity.name
        return entity.lower()  # block key is lower case

    @property
    def block_records(self) -> Table:
        return self.doc.block_records  # type: ignore

    @property
    def entitydb(self) -> EntityDB:
        return self.doc.entitydb  # type: ignore

    def load(self, entities: list[DXFEntity]) -> None:
        """
        Load DXF entities into BlockLayouts. `entities` is a list of
        entity tags, separated by BLOCK and ENDBLK entities.

        """

        def load_block_record(
            block: Block,
            endblk: EndBlk,
            block_entities: list[DXFEntity],
        ) -> BlockRecord | None:
            try:
                block_record = cast("BlockRecord", block_records.get(block.dxf.name))
            # Special case DXF R12 - has no BLOCK_RECORD table
            except DXFTableEntryError:
                block_record = cast(
                    "BlockRecord",
                    block_records.new(block.dxf.name, dxfattribs={"scale": 0}),
                )
            except DXFTypeError:
                raise DXFStructureError(
                    f"Invalid or missing name of BLOCK #{block.dxf.handle}"
                )
            # The BLOCK_RECORD is the central object which stores all the
            # information about a BLOCK and also owns all the entities of
            # this block definition.
            block_record.set_block(block, endblk)
            for entity in block_entities:
                block_record.add_entity(entity)  # type: ignore
            return block_record

        def link_entities() -> Iterable["DXFEntity"]:
            linked = entity_linker()
            for entity in entities:
                # Do not store linked entities (VERTEX, ATTRIB, SEQEND) in
                # the block layout, linked entities ares stored in their
                # parent entity e.g. VERTEX -> POLYLINE:
                if not linked(entity):
                    yield entity

        block_records = self.block_records
        section_head = cast("DXFTagStorage", entities[0])
        if section_head.dxftype() != "SECTION" or section_head.base_class[1] != (
            2,
            "BLOCKS",
        ):
            raise DXFStructureError("Critical structure error in BLOCKS section.")
        # Remove SECTION entity
        del entities[0]
        content: list[DXFEntity] = []
        block: Block = _MISSING_BLOCK_
        for entity in link_entities():
            if isinstance(entity, Block):
                if block is not _MISSING_BLOCK_:
                    logger.warning("Missing required ENDBLK, ignoring content.")
                block = entity
                content.clear()
            elif isinstance(entity, EndBlk):
                if block is _MISSING_BLOCK_:
                    logger.warning(
                        "Found ENDBLK without a preceding BLOCK, ignoring content."
                    )
                else:
                    block_name = block.dxf.get("name", "")
                    handle = block.dxf.get("handle", "<undefined>")
                    if not block_name:
                        block_name = recover_block_name(block)
                        if block_name:
                            logger.info(
                                f'Recovered block name "{block_name}" for block #{handle}.'
                            )
                            block.dxf.name = block_name
                    if block_name:
                        block_record = load_block_record(block, entity, content)
                        if isinstance(block_record, BlockRecord):
                            self.add(block_record)
                        else:
                            logger.warning(
                                f"Ignoring invalid BLOCK definition #{handle}."
                            )
                    else:
                        logger.warning(f"Ignoring BLOCK without name #{handle}.")
                    block = _MISSING_BLOCK_
                content.clear()
            else:
                # No check for valid entities here:
                # Use the audit or the recover module to fix invalid DXF files!
                content.append(entity)

    def _reconstruct_orphaned_block_records(self):
        """Find BLOCK_RECORD entries without block definition in the blocks
        section and create block definitions for this orphaned block records.

        """
        for block_record in self.block_records:
            if block_record.block is None:
                block = factory.create_db_entry(
                    "BLOCK",
                    dxfattribs={
                        "name": block_record.dxf.name,
                        "base_point": (0, 0, 0),
                    },
                    doc=self.doc,
                )
                endblk = factory.create_db_entry(
                    "ENDBLK",
                    dxfattribs={},
                    doc=self.doc,
                )
                block_record.set_block(block, endblk)
                self.add(block_record)

    def export_dxf(self, tagwriter: AbstractTagWriter) -> None:
        tagwriter.write_str("  0\nSECTION\n  2\nBLOCKS\n")
        for block_record in self.block_records:
            assert isinstance(block_record, BlockRecord)
            block_record.export_block_definition(tagwriter)
        tagwriter.write_tag2(0, "ENDSEC")

    def add(self, block_record: BlockRecord) -> BlockLayout:
        """Add or replace a block layout object defined by its block record.
        (internal API)
        """
        block_layout = BlockLayout(block_record)
        block_record.block_layout = block_layout
        assert self.block_records.has_entry(block_record.dxf.name)
        return block_layout

    def __iter__(self) -> Iterator[BlockLayout]:
        """Iterable of all :class:`~ezdxf.layouts.BlockLayout` objects."""
        return (block_record.block_layout for block_record in self.block_records)

    def __contains__(self, name: str) -> bool:
        """Returns ``True`` if :class:`~ezdxf.layouts.BlockLayout` `name`
        exist.
        """
        return self.block_records.has_entry(name)

    def __getitem__(self, name: str) -> BlockLayout:
        """Returns :class:`~ezdxf.layouts.BlockLayout` `name`,
        raises :class:`DXFKeyError` if `name` not exist.
        """
        try:
            block_record = cast("BlockRecord", self.block_records.get(name))
            return block_record.block_layout  # type: ignore
        except DXFTableEntryError:
            raise DXFKeyError(name)

    def __delitem__(self, name: str) -> None:
        """Deletes :class:`~ezdxf.layouts.BlockLayout` `name` and all of
        its content, raises :class:`DXFKeyError` if `name` not exist.
        """
        if name in self:
            self.block_records.remove(name)
        else:
            raise DXFKeyError(f'Block "{name}" does not exist.')

    def block_names(self) -> list[str]:
        """Returns a list of all block names."""
        return list(self.doc.block_records.entries.keys())  # type: ignore

    def get(self, name: str, default=None) -> BlockLayout:
        """Returns :class:`~ezdxf.layouts.BlockLayout` `name`, returns
        `default` if `name` not exist.
        """
        try:
            return self.__getitem__(name)
        except DXFKeyError:
            return default

    def get_block_layout_by_handle(self, block_record_handle: str) -> BlockLayout:
        """Returns a block layout by block record handle. (internal API)"""
        return self.doc.entitydb[block_record_handle].block_layout  # type: ignore

    def new(
        self,
        name: str,
        base_point: UVec = NULLVEC,
        dxfattribs=None,
    ) -> BlockLayout:
        """Create and add a new :class:`~ezdxf.layouts.BlockLayout`, `name`
        is the BLOCK name, `base_point` is the insertion point of the BLOCK.
        """
        assert self.doc is not None
        block_record = cast(BlockRecord, self.doc.block_records.new(name))

        dxfattribs = dxfattribs or {}
        dxfattribs["owner"] = block_record.dxf.handle
        dxfattribs["name"] = name
        dxfattribs["base_point"] = Vec3(base_point)
        head = factory.create_db_entry("BLOCK", dxfattribs, self.doc)
        tail = factory.create_db_entry(
            "ENDBLK", {"owner": block_record.dxf.handle}, doc=self.doc
        )
        block_record.set_block(head, tail)  # type: ignore
        return self.add(block_record)

    def new_anonymous_block(
        self, type_char: str = "U", base_point: UVec = NULLVEC
    ) -> BlockLayout:
        """Create and add a new anonymous :class:`~ezdxf.layouts.BlockLayout`,
        `type_char` is the BLOCK type, `base_point` is the insertion point of
        the BLOCK.

            ========= ==========
            type_char Anonymous Block Type
            ========= ==========
            ``'U'``   ``'*U###'`` anonymous BLOCK
            ``'E'``   ``'*E###'`` anonymous non-uniformly scaled BLOCK
            ``'X'``   ``'*X###'`` anonymous HATCH graphic
            ``'D'``   ``'*D###'`` anonymous DIMENSION graphic
            ``'A'``   ``'*A###'`` anonymous GROUP
            ``'T'``   ``'*T###'`` anonymous block for ACAD_TABLE content
            ========= ==========

        """
        block_name = self.anonymous_block_name(type_char)
        block = self.new(block_name, base_point, {"flags": const.BLK_ANONYMOUS})
        return block

    def anonymous_block_name(self, type_char: str) -> str:
        """Create name for an anonymous block. (internal API)

        Args:
            type_char: letter

                U = *U### anonymous blocks
                E = *E### anonymous non-uniformly scaled blocks
                X = *X### anonymous hatches
                D = *D### anonymous dimensions
                A = *A### anonymous groups
                T = *T### anonymous ACAD_TABLE content

        """
        while True:
            self._anonymous_block_counter += 1
            block_name = f"*{type_char}{self._anonymous_block_counter}"
            if not self.block_records.has_entry(block_name):
                return block_name

    def rename_block(self, old_name: str, new_name: str) -> None:
        """Rename :class:`~ezdxf.layouts.BlockLayout` `old_name` to `new_name`

        .. warning::

            This is a low-level tool and does not rename the block references,
            so all block references to `old_name` are pointing to a non-existing
            block definition!

        """
        block_record = cast(BlockRecord, self.block_records.get(old_name))
        block_record.rename(new_name)
        self.block_records.replace(old_name, block_record)
        self.add(block_record)

    def delete_block(self, name: str, safe: bool = True) -> None:
        """Delete block.

        Applies some safety checks when `safe` is ``True``.
        A :class:`DXFBlockInUseError` will be raised for:

            - blocks with active references
            - blocks representing existing layouts
            - special blocks used internally

        Args:
            name: block name (case-insensitive)
            safe: apply safety checks

        Raises:
            DXFKeyError: if block not exists
            DXFBlockInUseError: when safe is ``True`` and block is in use
        """
        if safe:
            assert self.doc is not None, "valid DXF document required"
            block = self.doc.blocks.get(name)
            if block is None:
                raise DXFKeyError(f'Block "{name}" does not exist.')
            if not block.is_alive:
                return  # block is already destroyed
            if block.is_any_layout:
                raise DXFBlockInUseError(
                    f'Block "{name}" represents an existing layout.'
                )
            if is_special_block(name):
                raise DXFBlockInUseError(
                    f'Special block "{name}" maybe used without explicit INSERT entity.'
                )
            query_string = f'INSERT[name=="{name}"]i'
            try:
                block_refs = self.doc.query(query_string)  # ignore case
            except ParseException:
                logger.error(f'Parsing error in query string: "{query_string}"')
                return

            if len(block_refs):
                raise DXFBlockInUseError(f'Block "{name}" is still in use.')
        self.__delitem__(name)

    def delete_all_blocks(self) -> None:
        """Delete all blocks without references except modelspace- or
        paperspace layout blocks, special arrow- and anonymous blocks
        (DIMENSION, ACAD_TABLE).

        .. warning::

            There could exist references to blocks which are not documented in the DXF
            reference, hidden in extended data sections or application defined data,
            which could invalidate a DXF document if these blocks will be deleted.

        """
        assert self.doc is not None
        active_references = set(
            validator.make_table_key(entity.dxf.name)
            for entity in self.doc.query("INSERT")
        )

        def is_safe(name: str) -> bool:
            if is_special_block(name):
                return False
            return name not in active_references

        trash = set()
        for block in self:
            name = validator.make_table_key(block.name)
            if not block.is_any_layout and is_safe(name):
                trash.add(name)

        for name in trash:
            self.__delitem__(name)

    def audit(self, auditor: Auditor) -> None:
        """Audit and repair BLOCKS section.

        .. important::

            Do not delete entities during the auditing process as this will alter
            the entity database while iterating it, instead use::

                auditor.trash(entity)

            to delete invalid entities after auditing automatically.

        """
        assert self.doc is auditor.doc, "Auditor for different DXF document."

        for block_record in self.block_records:
            assert isinstance(block_record, BlockRecord)

            block_record_handle: str = block_record.dxf.handle
            unlink_entities: list[DXFEntity] = []
            es = block_record.entity_space
            for entity in es:
                if not is_graphic_entity(entity):
                    auditor.fixed_error(
                        code=AuditError.REMOVED_INVALID_GRAPHIC_ENTITY,
                        message=f"Removed invalid DXF entity {str(entity)} from"
                        f" BLOCK '{block_record.dxf.name}'.",
                    )
                    auditor.trash(entity)
                elif isinstance(entity, Attrib):
                    # ATTRIB can only exist as an attached entity of the INSERT
                    # entity!
                    auditor.fixed_error(
                        code=AuditError.REMOVED_STANDALONE_ATTRIB_ENTITY,
                        message=f"Removed standalone {str(entity)} entity from"
                        f" BLOCK '{block_record.dxf.name}'.",
                    )
                    auditor.trash(entity)

                if not entity.is_alive:
                    continue

                if entity.dxf.owner != block_record_handle:
                    auditor.fixed_error(
                        code=AuditError.REMOVED_ENTITY_WITH_INVALID_OWNER_HANDLE,
                        message=f"Removed DXF entity {str(entity)} with invalid owner "
                        f"handle (#{entity.dxf.owner} != #{block_record_handle}) "
                        f"from BLOCK '{block_record.dxf.name}'.",
                    )
                    # do not destroy the entity, it's maybe owned to another block
                    unlink_entities.append(entity)

            for entity in unlink_entities:
                if entity.is_alive:
                    try:
                        es.remove(entity)
                    except ValueError:
                        pass



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.entities.sun import Sun
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

SUN = """0
SUN
5
0
330
0
100
AcDbSun
90
1
290
1
63
7
421
16777215
40
1.0
291
1
91
2456922
92
43200
292
0
70
0
71
256
280
1
"""


@pytest.fixture
def entity():
    return Sun.from_text(SUN)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "SUN" in ENTITY_CLASSES


def test_default_init():
    entity = Sun()
    assert entity.dxftype() == "SUN"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = Sun.new(handle="ABBA", owner="0", dxfattribs={})
    assert entity.dxf.version == 1
    assert entity.dxf.status == 1
    assert entity.dxf.color == 7
    assert entity.dxf.true_color == 16777215
    assert entity.dxf.intensity == 1
    assert entity.dxf.shadows == 1
    assert entity.dxf.julian_day == 2456922
    assert entity.dxf.time == 43200
    assert entity.dxf.daylight_savings_time == 0
    assert entity.dxf.shadow_type == 0
    assert entity.dxf.shadow_map_size == 256
    assert entity.dxf.shadow_softness == 1


def test_load_from_text(entity):
    assert entity.dxf.version == 1
    assert entity.dxf.status == 1
    assert entity.dxf.color == 7
    assert entity.dxf.true_color == 16777215
    assert entity.dxf.intensity == 1
    assert entity.dxf.shadows == 1
    assert entity.dxf.julian_day == 2456922
    assert entity.dxf.time == 43200
    assert entity.dxf.daylight_savings_time == 0
    assert entity.dxf.shadow_type == 0
    assert entity.dxf.shadow_map_size == 256
    assert entity.dxf.shadow_softness == 1


def test_write_dxf():
    entity = Sun.from_text(SUN)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(SUN)
    assert result == expected


def test_sun():
    doc = ezdxf.new("R2007")
    sun = doc.objects.new_entity("SUN", {})
    assert sun.dxftype() == "SUN"
    assert sun.dxf.version == 1



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf

CONST_GUID = "{00000000-0000-0000-0000-000000000000}"
FINGERPRINT_GUID = "$FINGERPRINTGUID"
VERSION_GUID = "$VERSIONGUID"


def test_guids_are_created_for_new_docs():
    doc = ezdxf.new()
    assert doc.header[FINGERPRINT_GUID] != CONST_GUID
    assert doc.header[VERSION_GUID] != CONST_GUID


def test_guids_are_preserved_at_loading(tmp_path):
    filepath = tmp_path / "doc1.dxf"
    doc = ezdxf.new()
    doc.saveas(filepath)
    doc1 = ezdxf.readfile(filepath)
    assert doc1.header[FINGERPRINT_GUID] == doc.header[FINGERPRINT_GUID]
    assert doc1.header[VERSION_GUID] == doc.header[VERSION_GUID]


def test_fingerprint_guid_is_preserved_at_export(tmp_path):
    filepath1 = tmp_path / "doc1.dxf"
    filepath2 = tmp_path / "doc2.dxf"
    doc = ezdxf.new()
    doc.saveas(filepath1)
    doc1 = ezdxf.readfile(filepath1)
    doc1.saveas(filepath2)
    doc2 = ezdxf.readfile(filepath2)
    assert doc1.header[FINGERPRINT_GUID] == doc.header[FINGERPRINT_GUID]
    assert doc1.header[FINGERPRINT_GUID] == doc2.header[FINGERPRINT_GUID]


def test_version_guid_changes_at_export(tmp_path):
    filepath1 = tmp_path / "doc1.dxf"
    filepath2 = tmp_path / "doc2.dxf"
    doc = ezdxf.new()
    doc.saveas(filepath1)
    doc1 = ezdxf.readfile(filepath1)
    assert doc1.header[VERSION_GUID] == doc.header[VERSION_GUID]
    # Exporting a new version of this document, $VERSIONGUID must change.
    old_version_guid = doc1.header[VERSION_GUID]
    doc1.saveas(filepath2)  # changes $VERSIONGUID of doc1!
    assert old_version_guid != doc1.header[VERSION_GUID]
    doc2 = ezdxf.readfile(filepath2)
    assert old_version_guid != doc2.header[VERSION_GUID]
    assert doc1.header[VERSION_GUID] == doc2.header[VERSION_GUID]


if __name__ == '__main__':
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.entities.mtext import (
    load_columns_from_embedded_object,
    MText,
    ColumnType,
    MTextColumns,
)
from ezdxf.lldxf import const
from ezdxf.lldxf.types import EMBEDDED_OBJ_STR, EMBEDDED_OBJ_MARKER
from ezdxf.lldxf.tags import Tags
from ezdxf.lldxf.tagwriter import TagCollector

DYNAMIC_MANUAL_HEIGHT = """101
Embedded Object
70
1
10
1.0
20
0.0
30
0.0
11
69.8
21
276.1
31
0.0
40
62.6
41
0.0
42
175.0
43
164.8
71
2
72
3
44
50.0
45
12.5
73
0
74
0
46
164.8
46
154.3
46
0.0
"""

DYNAMIC_AUTO_HEIGHT = """101
Embedded Object
70
1
10
1.0
20
0.0
30
0.0
11
69.8
21
276.1
31
0.0
40
62.6
41
158.1
42
175.0
43
158.1
71
2
72
0
44
50.0
45
12.5
73
1
74
0
"""

STATIC = """101
Embedded Object
70
1
10
1.0
20
0.0
30
0.0
11
69.8
21
276.1
31
0.0
40
62.6
41
150.0
42
175.0
43
150.0
71
1
72
3
44
50.0
45
12.5
73
0
74
0
"""


# The dynamic auto height and static types are very similar, the difference is
# important for CAD applications, but not for the DXF format itself.


@pytest.mark.parametrize(
    "obj",
    [DYNAMIC_MANUAL_HEIGHT, DYNAMIC_AUTO_HEIGHT, STATIC],
    ids=["DYN_MANUAL", "DYN_AUTO", "STATIC"],
)
def test_load_mtext_attribs_from_embedded_object(obj):
    embedded_obj = Tags.from_text(obj)
    dxf = MText().dxf
    dxf.rotation = 45

    load_columns_from_embedded_object(dxf, embedded_obj)
    assert dxf.width == 62.6
    assert dxf.text_direction == (1, 0, 0)
    assert dxf.hasattr("rotation") is False, "remove rotation attribute"
    assert dxf.insert == (69.8, 276.1, 0)


def test_load_dynamic_cols_manual_height():
    """Every column can have a different height."""
    embedded_obj = Tags.from_text(DYNAMIC_MANUAL_HEIGHT)
    cols = load_columns_from_embedded_object(MText().dxf, embedded_obj)
    assert cols.count == 3
    assert cols.column_type == ColumnType.DYNAMIC
    assert cols.auto_height is False
    assert cols.reversed_column_flow is False
    assert cols.defined_height == 0.0, "not defined if auto_height is False"
    assert cols.width == 50.0
    assert cols.gutter_width == 12.5
    assert cols.total_width == 175.0  # 3 * 50 + 2 * 12.5
    # total_height = max(heights) even if the last column is the tallest
    assert cols.total_height == 164.8
    assert cols.total_height == max(cols.heights)
    assert len(cols.linked_columns) == 0, "MTEXT is a single entity in R2018"
    assert len(cols.heights) == 3
    assert cols.heights[-1] == 0.0, "last column height has to be 0.0"
    assert cols.heights == [164.8, 154.3, 0.0]


def test_load_dynamic_cols_with_auto_height():
    """All columns have the same column height."""
    embedded_obj = Tags.from_text(DYNAMIC_AUTO_HEIGHT)
    cols = load_columns_from_embedded_object(MText().dxf, embedded_obj)
    # Count is a calculated value, group code 72 (column height count) is 0!
    assert cols.count == 3
    assert cols.column_type == ColumnType.DYNAMIC
    assert cols.auto_height is True
    assert cols.reversed_column_flow is False
    assert cols.defined_height == 158.1, "required if auto_height is True"
    assert cols.width == 50.0
    assert cols.gutter_width == 12.5
    assert cols.total_width == 175.0  # 3 * 50 + 2 * 12.5
    assert cols.total_height == 158.1
    assert len(cols.linked_columns) == 0, "MTEXT is a single entity in R2018"
    assert len(cols.heights) == 0


def test_load_static_cols():
    """All columns have the same column height."""
    embedded_obj = Tags.from_text(STATIC)
    cols = load_columns_from_embedded_object(MText().dxf, embedded_obj)
    assert cols.count == 3
    assert cols.column_type == ColumnType.STATIC
    assert cols.auto_height is False
    assert cols.reversed_column_flow is False
    assert cols.defined_height == 150.0, "required for static columns"
    assert cols.width == 50.0
    assert cols.gutter_width == 12.5
    assert cols.total_width == 175.0  # 3 * 50 + 2 * 12.5
    assert cols.total_height == 150.0
    assert len(cols.linked_columns) == 0, "MTEXT is a single entity in R2018"
    assert len(cols.heights) == 0


def make_mtext(txt: str):
    mtext = MText()
    embedded_obj = Tags.from_text(txt)
    mtext._columns = load_columns_from_embedded_object(mtext.dxf, embedded_obj)
    return mtext


def test_export_static_columns_as_embedded_object():
    mtext = make_mtext(STATIC)
    collector = TagCollector(dxfversion=const.DXF2018)
    mtext.export_embedded_object(collector)
    tags = collector.tags
    assert len(tags) == 18
    assert tags[0] == (EMBEDDED_OBJ_MARKER, EMBEDDED_OBJ_STR)
    assert tags[1] == (70, 1)
    assert tags[2] == (10, 1)
    assert tags[3] == (20, 0)
    assert tags[4] == (30, 0)
    assert tags[5] == (11, 69.8)
    assert tags[6] == (21, 276.1)
    assert tags[7] == (31, 0)
    assert tags[8] == (40, 62.6)
    assert tags[9] == (41, 150.0)
    assert tags[10] == (42, 175.0)
    assert tags[11] == (43, 150.0)
    assert tags[12] == (71, 1)
    assert tags[13] == (72, 3)
    assert tags[14] == (44, 50.0)
    assert tags[15] == (45, 12.5)
    assert tags[16] == (73, 0)
    assert tags[17] == (74, 0)


def test_export_dynamic_columns_auto_height_as_embedded_object():
    mtext = make_mtext(DYNAMIC_AUTO_HEIGHT)
    collector = TagCollector(dxfversion=const.DXF2018)
    mtext.export_embedded_object(collector)
    tags = collector.tags
    assert len(tags) == 18
    assert tags[0] == (EMBEDDED_OBJ_MARKER, EMBEDDED_OBJ_STR)
    assert tags[1] == (70, 1)
    assert tags[2] == (10, 1)
    assert tags[3] == (20, 0)
    assert tags[4] == (30, 0)
    assert tags[5] == (11, 69.8)
    assert tags[6] == (21, 276.1)
    assert tags[7] == (31, 0)
    assert tags[8] == (40, 62.6)
    assert tags[9] == (41, 158.1)
    assert tags[10] == (42, 175.0)
    assert tags[11] == (43, 158.1)
    assert tags[12] == (71, 2)
    assert tags[13] == (72, 0)
    assert tags[14] == (44, 50.0)
    assert tags[15] == (45, 12.5)
    assert tags[16] == (73, 1)
    assert tags[17] == (74, 0)


def test_export_dynamic_columns_manual_height_as_embedded_object():
    mtext = make_mtext(DYNAMIC_MANUAL_HEIGHT)
    collector = TagCollector(dxfversion=const.DXF2018)
    mtext.export_embedded_object(collector)
    tags = collector.tags
    assert len(tags) == 21
    assert tags[0] == (EMBEDDED_OBJ_MARKER, EMBEDDED_OBJ_STR)
    assert tags[1] == (70, 1)
    assert tags[2] == (10, 1)
    assert tags[3] == (20, 0)
    assert tags[4] == (30, 0)
    assert tags[5] == (11, 69.8)
    assert tags[6] == (21, 276.1)
    assert tags[7] == (31, 0)
    assert tags[8] == (40, 62.6)
    assert tags[9] == (41, 0.0)
    assert tags[10] == (42, 175.0)
    assert tags[11] == (43, 164.8)
    assert tags[12] == (71, 2)
    assert tags[13] == (72, 3)
    assert tags[14] == (44, 50.0)
    assert tags[15] == (45, 12.5)
    assert tags[16] == (73, 0)
    assert tags[17] == (74, 0)
    assert tags[18] == (46, 164.8)
    assert tags[19] == (46, 154.3)
    assert tags[20] == (46, 0)


def new_mtext_with_columns(count=3):
    columns = MTextColumns()
    columns.count = count
    columns.width = 10
    columns.gutter_width = 0.5
    columns.defined_height = 50
    mtext = MText.new()
    mtext.setup_columns(columns, linked=False)
    return mtext


def test_create_new_mtext_with_columns():
    mtext = new_mtext_with_columns(3)
    columns = mtext.columns
    assert columns.column_type == ColumnType.STATIC
    assert len(columns.linked_columns) == 0
    assert len(columns.heights) == 0, "all columns have the same defined height"
    assert mtext.dxf.width == columns.width
    assert columns.total_height == columns.defined_height
    assert columns.total_width == 31

    # default location and text direction:
    assert mtext.dxf.insert == (0, 0, 0)
    assert mtext.dxf.text_direction == (1, 0, 0)


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2024 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Iterable,
    Iterator,
    cast,
    Union,
    Optional,
    Callable,
)
from typing_extensions import Self
import math
import logging

from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
    merge_group_code_mappings,
)
from ezdxf.lldxf.const import (
    DXF12,
    SUBCLASS_MARKER,
    DXFValueError,
    DXFKeyError,
    DXFStructureError,
)
from ezdxf.math import (
    Vec3,
    UVec,
    X_AXIS,
    Y_AXIS,
    Z_AXIS,
    Matrix44,
    UCS,
    NULLVEC,
)
from ezdxf.math.transformtools import (
    InsertTransformationError,
    InsertCoordinateSystem,
)
from ezdxf.explode import (
    explode_block_reference,
    virtual_block_reference_entities,
)
from ezdxf.entities import factory
from ezdxf.query import EntityQuery
from ezdxf.audit import AuditError
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import (
    DXFGraphic,
    acdb_entity,
    elevation_to_z_axis,
    acdb_entity_group_codes,
)
from .subentity import LinkedEntities
from .attrib import Attrib

if TYPE_CHECKING:
    from ezdxf.audit import Auditor
    from ezdxf.entities import DXFNamespace, AttDef, DXFEntity
    from ezdxf.layouts import BaseLayout, BlockLayout
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf import xref

__all__ = ["Insert"]

logger = logging.getLogger("ezdxf")
ABS_TOL = 1e-9

# DXF files as XREF:
# The INSERT entity is used to attach XREFS.
# The model space is the block content, if the whole document is used as
# BLOCK (XREF), but this is only supported for the DWG format.
# AutoCAD does not support DXF files as XREFS, they are ignored, but the DXF
# file is valid! BricsCAD shows DXF files as XREFS, but does not allow to attach
# DXF files as XREFS by the application itself.

# Multi-INSERT has subclass id AcDbMInsertBlock
acdb_block_reference = DefSubclass(
    "AcDbBlockReference",
    {
        "attribs_follow": DXFAttr(66, default=0, optional=True),
        "name": DXFAttr(2, validator=validator.is_valid_block_name),
        "insert": DXFAttr(10, xtype=XType.any_point),
        # Elevation is a legacy feature from R11 and prior, do not use this
        # attribute, store the entity elevation in the z-axis of the vertices.
        # ezdxf does not export the elevation attribute!
        "elevation": DXFAttr(38, default=0, optional=True),
        "xscale": DXFAttr(
            41,
            default=1,
            optional=True,
            validator=validator.is_not_zero,
            fixer=RETURN_DEFAULT,
        ),
        "yscale": DXFAttr(
            42,
            default=1,
            optional=True,
            validator=validator.is_not_zero,
            fixer=RETURN_DEFAULT,
        ),
        "zscale": DXFAttr(
            43,
            default=1,
            optional=True,
            validator=validator.is_not_zero,
            fixer=RETURN_DEFAULT,
        ),
        "rotation": DXFAttr(50, default=0, optional=True),
        "column_count": DXFAttr(
            70,
            default=1,
            optional=True,
            validator=validator.is_greater_zero,
            fixer=RETURN_DEFAULT,
        ),
        "row_count": DXFAttr(
            71,
            default=1,
            optional=True,
            validator=validator.is_greater_zero,
            fixer=RETURN_DEFAULT,
        ),
        "column_spacing": DXFAttr(44, default=0, optional=True),
        "row_spacing": DXFAttr(45, default=0, optional=True),
        "extrusion": DXFAttr(
            210,
            xtype=XType.point3d,
            default=Z_AXIS,
            optional=True,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
    },
)
acdb_block_reference_group_codes = group_code_mapping(acdb_block_reference)
merged_insert_group_codes = merge_group_code_mappings(
    acdb_entity_group_codes, acdb_block_reference_group_codes  # type: ignore
)


# Notes to SEQEND:
#
# The INSERT entity requires only a SEQEND if ATTRIB entities are attached.
#  So a loaded INSERT could have a missing SEQEND.
#
# A bounded INSERT needs a SEQEND to be valid at export if there are attached
# ATTRIB entities, but the LinkedEntities.post_bind_hook() method creates
# always a new SEQEND after binding the INSERT entity to a document.
#
# Nonetheless, the Insert.add_attrib() method also creates the required SEQEND entity if
# necessary.


@factory.register_entity
class Insert(LinkedEntities):
    """DXF INSERT entity

    The INSERT entity is hard owner of its ATTRIB entities and the SEQEND entity:

        ATTRIB.dxf.owner == INSERT.dxf.handle
        SEQEND.dxf.owner == INSERT.dxf.handle

    Note:

        The ATTDEF entity in block definitions is owned by the BLOCK_RECORD like
        all graphical entities.

    """

    DXFTYPE = "INSERT"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_block_reference)

    @property
    def attribs(self) -> list[Attrib]:
        return self._sub_entities  # type: ignore

    @property
    def attribs_follow(self) -> bool:
        return bool(len(self.attribs))

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        """Loading interface. (internal API)"""
        # bypass DXFGraphic, loading proxy graphic is skipped!
        dxf = super(DXFGraphic, self).load_dxf_attribs(processor)
        if processor:
            processor.simple_dxfattribs_loader(dxf, merged_insert_group_codes)
            if processor.r12:
                # Transform elevation attribute from R11 to z-axis values:
                elevation_to_z_axis(dxf, ("insert",))
        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        if tagwriter.dxfversion > DXF12:
            if (self.dxf.column_count > 1) or (self.dxf.row_count > 1):
                tagwriter.write_tag2(SUBCLASS_MARKER, "AcDbMInsertBlock")
            else:
                tagwriter.write_tag2(SUBCLASS_MARKER, "AcDbBlockReference")
        if self.attribs_follow:
            tagwriter.write_tag2(66, 1)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "name",
                "insert",
                "xscale",
                "yscale",
                "zscale",
                "rotation",
                "column_count",
                "row_count",
                "column_spacing",
                "row_spacing",
                "extrusion",
            ],
        )

    def export_dxf(self, tagwriter: AbstractTagWriter):
        super().export_dxf(tagwriter)
        # Do no export SEQEND if no ATTRIBS attached:
        if self.attribs_follow:
            self.process_sub_entities(lambda e: e.export_dxf(tagwriter))

    def register_resources(self, registry: xref.Registry) -> None:
        # The attached ATTRIB entities are registered by the parent class LinkedEntities
        super().register_resources(registry)
        registry.add_block_name(self.dxf.name)

    def map_resources(self, clone: Self, mapping: xref.ResourceMapper) -> None:
        # The attached ATTRIB entities are mapped by the parent class LinkedEntities
        super().map_resources(clone, mapping)
        clone.dxf.name = mapping.get_block_name(self.dxf.name)

    @property
    def has_scaling(self) -> bool:
        """Returns ``True`` if scaling is applied to any axis."""
        if self.dxf.hasattr("xscale") and self.dxf.xscale != 1:
            return True
        if self.dxf.hasattr("yscale") and self.dxf.yscale != 1:
            return True
        if self.dxf.hasattr("zscale") and self.dxf.zscale != 1:
            return True
        return False

    @property
    def has_uniform_scaling(self) -> bool:
        """Returns ``True`` if the scale factor is uniform for x-, y- and z-axis,
        ignoring reflections e.g. (1, 1, -1) is uniform scaling.

        """
        return abs(self.dxf.xscale) == abs(self.dxf.yscale) == abs(self.dxf.zscale)

    def set_scale(self, factor: float):
        """Set a uniform scale factor."""
        if factor == 0:
            raise ValueError("Invalid scale factor.")
        self.dxf.xscale = factor
        self.dxf.yscale = factor
        self.dxf.zscale = factor
        return self

    def block(self) -> Optional[BlockLayout]:
        """Returns the associated :class:`~ezdxf.layouts.BlockLayout`."""
        if self.doc:
            return self.doc.blocks.get(self.dxf.name)
        return None

    def is_xref(self) -> bool:
        """Return ``True`` if the INSERT entity represents a XREF or XREF_OVERLAY."""
        block = self.block()
        if block is not None:
            return block.block_record.is_xref
        return False

    def place(
        self,
        insert: Optional[UVec] = None,
        scale: Optional[tuple[float, float, float]] = None,
        rotation: Optional[float] = None,
    ) -> Insert:
        """
        Set the location, scaling and rotation attributes. Arguments which are ``None``
        will be ignored.

        Args:
            insert: insert location as (x, y [,z]) tuple
            scale: (x-scale, y-scale, z-scale) tuple
            rotation : rotation angle in degrees

        """
        if insert is not None:
            self.dxf.insert = insert
        if scale is not None:
            if len(scale) != 3:
                raise DXFValueError("Argument scale has to be a (x, y[, z]) tuple.")
            x, y, z = scale
            self.dxf.xscale = x
            self.dxf.yscale = y
            self.dxf.zscale = z
        if rotation is not None:
            self.dxf.rotation = rotation
        return self

    def grid(
        self, size: tuple[int, int] = (1, 1), spacing: tuple[float, float] = (1, 1)
    ) -> Insert:
        """Place block reference in a grid layout, grid `size` defines the
        row- and column count, `spacing` defines the distance between two block
        references.

        Args:
            size: grid size as (row_count, column_count) tuple
            spacing: distance between placing as (row_spacing, column_spacing) tuple

        """
        try:
            rows, cols = size
        except ValueError:
            raise DXFValueError("Size has to be a (row_count, column_count) tuple.")
        self.dxf.row_count = rows
        self.dxf.column_count = cols
        try:
            row_spacing, col_spacing = spacing
        except ValueError:
            raise DXFValueError(
                "Spacing has to be a (row_spacing, column_spacing) tuple."
            )
        self.dxf.row_spacing = row_spacing
        self.dxf.column_spacing = col_spacing
        return self

    def get_attrib(
        self, tag: str, search_const: bool = False
    ) -> Optional[Union[Attrib, AttDef]]:
        """Get an attached :class:`Attrib` entity with the given `tag`,
        returns ``None`` if not found.  Some applications do not attach constant
        ATTRIB entities, set `search_const` to ``True``, to get at least the
        associated :class:`AttDef` entity.

        Args:
            tag: tag name of the ATTRIB entity
            search_const: search also const ATTDEF entities

        """
        for attrib in self.attribs:
            if tag == attrib.dxf.tag:
                return attrib
        if search_const and self.doc is not None:
            block = self.doc.blocks[self.dxf.name]
            for attdef in block.get_const_attdefs():
                if tag == attdef.dxf.tag:
                    return attdef
        return None

    def get_attrib_text(
        self, tag: str, default: str = "", search_const: bool = False
    ) -> str:
        """Get content text of an attached :class:`Attrib` entity with
        the given `tag`, returns the `default` value if not found.
        Some applications do not attach constant ATTRIB entities, set
        `search_const` to ``True``, to get content text of the
        associated :class:`AttDef` entity.

        Args:
            tag: tag name of the ATTRIB entity
            default: default value if ATTRIB `tag` is absent
            search_const: search also const ATTDEF entities

        """
        attrib = self.get_attrib(tag, search_const)
        if attrib is None:
            return default
        return attrib.dxf.text

    def has_attrib(self, tag: str, search_const: bool = False) -> bool:
        """Returns ``True`` if the INSERT entity has an attached ATTRIB entity with the
        given `tag`.  Some applications do not attach constant ATTRIB entities, set
        `search_const` to ``True``, to check for an associated :class:`AttDef` entity
        with constant content.


        Args:
            tag: tag name fo the ATTRIB entity
            search_const: search also const ATTDEF entities

        """
        return self.get_attrib(tag, search_const) is not None

    def add_attrib(
        self, tag: str, text: str, insert: UVec = (0, 0), dxfattribs=None
    ) -> Attrib:
        """Attach an :class:`Attrib` entity to the block reference.

        Example for appending an attribute to an INSERT entity::

            e.add_attrib('EXAMPLETAG', 'example text').set_placement(
                (3, 7), align=TextEntityAlignment.MIDDLE_CENTER
            )

        Args:
            tag: tag name of the ATTRIB entity
            text: content text as string
            insert: insert location as (x, y[, z]) tuple in :ref:`OCS`
            dxfattribs: additional DXF attributes for the ATTRIB entity

        """
        dxfattribs = dict(dxfattribs or {})
        dxfattribs["tag"] = tag
        dxfattribs["text"] = text
        dxfattribs["insert"] = insert
        attrib = cast("Attrib", self._new_compound_entity("ATTRIB", dxfattribs))
        self.attribs.append(attrib)

        # This case is only possible if the INSERT was read from a file without
        # attached ATTRIB entities:
        if self.seqend is None:
            self.new_seqend()
        return attrib

    def delete_attrib(self, tag: str, ignore=False) -> None:
        """Delete an attached :class:`Attrib` entity from INSERT. Raises an
        :class:`DXFKeyError` exception, if no ATTRIB for the given `tag` exist if
        `ignore` is ``False``.

        Args:
            tag: tag name of the ATTRIB entity
            ignore: ``False`` for raising :class:`DXFKeyError` if ATTRIB `tag`
                does not exist.

        Raises:
            DXFKeyError: no ATTRIB for the given `tag` exist

        """
        for index, attrib in enumerate(self.attribs):
            if attrib.dxf.tag == tag:
                del self.attribs[index]
                attrib.destroy()
                return
        if not ignore:
            raise DXFKeyError(tag)

    def delete_all_attribs(self) -> None:
        """Delete all :class:`Attrib` entities attached to the INSERT entity."""
        if not self.is_alive:
            return

        for attrib in self.attribs:
            attrib.destroy()
        self._sub_entities = []

    def transform(self, m: Matrix44) -> Insert:
        """Transform INSERT entity by transformation matrix `m` inplace.

        Unlike the transformation matrix `m`, the INSERT entity can not
        represent a non-orthogonal target coordinate system and an
        :class:`InsertTransformationError` will be raised in that case.

        """
        dxf = self.dxf
        source_system = InsertCoordinateSystem(
            insert=Vec3(dxf.insert),
            scale=(dxf.xscale, dxf.yscale, dxf.zscale),
            rotation=dxf.rotation,
            extrusion=dxf.extrusion,
        )
        try:
            target_system = source_system.transform(m, ABS_TOL)
        except InsertTransformationError:
            raise InsertTransformationError(
                "INSERT entity can not represent a non-orthogonal target coordinate system."
            )
        dxf.insert = target_system.insert
        dxf.rotation = target_system.rotation
        dxf.extrusion = target_system.extrusion
        dxf.xscale = target_system.scale_factor_x
        dxf.yscale = target_system.scale_factor_y
        dxf.zscale = target_system.scale_factor_z

        for attrib in self.attribs:
            attrib.transform(m)
        self.post_transform(m)
        return self

    def translate(self, dx: float, dy: float, dz: float) -> Insert:
        """Optimized INSERT translation about `dx` in x-axis, `dy` in y-axis
        and `dz` in z-axis.

        """
        ocs = self.ocs()
        self.dxf.insert = ocs.from_wcs(Vec3(dx, dy, dz) + ocs.to_wcs(self.dxf.insert))
        for attrib in self.attribs:
            attrib.translate(dx, dy, dz)
        return self

    def matrix44(self) -> Matrix44:
        """Returns a transformation matrix to transform the block entities from the
        block reference coordinate system into the :ref:`WCS`.
        """
        dxf = self.dxf
        sx = dxf.xscale
        sy = dxf.yscale
        sz = dxf.zscale

        ocs = self.ocs()
        extrusion = ocs.uz
        ux = Vec3(ocs.to_wcs(X_AXIS))
        uy = Vec3(ocs.to_wcs(Y_AXIS))
        m = Matrix44.ucs(ux=ux * sx, uy=uy * sy, uz=extrusion * sz)
        # same as Matrix44.ucs(ux, uy, extrusion) * Matrix44.scale(sx, sy, sz)

        angle = math.radians(dxf.rotation)
        if angle:
            m *= Matrix44.axis_rotate(extrusion, angle)

        insert = ocs.to_wcs(dxf.get("insert", NULLVEC))

        block_layout = self.block()
        if block_layout is not None:
            # transform block base point into WCS without translation
            insert -= m.transform_direction(block_layout.block.dxf.base_point)  # type: ignore

        # set translation
        m.set_row(3, insert.xyz)
        return m

    def ucs(self):
        """Returns the block reference coordinate system as :class:`ezdxf.math.UCS`
        object.
        """
        m = self.matrix44()
        ucs = UCS()
        ucs.matrix = m
        return ucs

    def reset_transformation(self) -> None:
        """Reset block reference attributes location, rotation angle and
        the extrusion vector but preserves the scale factors.

        """
        self.dxf.insert = NULLVEC
        self.dxf.discard("rotation")
        self.dxf.discard("extrusion")

    def explode(
        self, target_layout: Optional[BaseLayout] = None, *, redraw_order=False
    ) -> EntityQuery:
        """Explodes the block reference entities into the target layout, if target
        layout is ``None``, the layout of the block reference will be used.
        This method destroys the source block reference entity.

        Transforms the block entities into the required :ref:`WCS` location by
        applying the block reference attributes `insert`, `extrusion`,
        `rotation` and the scale factors `xscale`, `yscale` and `zscale`.

        Attached ATTRIB entities are converted to TEXT entities, this is the
        behavior of the BURST command of the AutoCAD Express Tools.

        .. warning::

            **Non-uniform scale factors** may lead to incorrect results some entities
            (TEXT, MTEXT, ATTRIB).

        Args:
            target_layout: target layout for exploded entities, ``None`` for
                same layout as source entity.
            redraw_order: create entities in ascending redraw order if ``True``

        Returns:
            :class:`~ezdxf.query.EntityQuery` container referencing all exploded
            DXF entities.

        """
        if target_layout is None:
            target_layout = self.get_layout()
            if target_layout is None:
                raise DXFStructureError(
                    "INSERT without layout assignment, specify target layout"
                )
        return explode_block_reference(
            self, target_layout=target_layout, redraw_order=redraw_order
        )

    def __virtual_entities__(self) -> Iterator[DXFGraphic]:
        """Implements the SupportsVirtualEntities protocol.

        This protocol is for consistent internal usage and does not replace
        the method :meth:`virtual_entities`! Ignores the redraw-order!
        """
        return self.virtual_entities()

    def virtual_entities(
        self,
        *,
        skipped_entity_callback: Optional[Callable[[DXFGraphic, str], None]] = None,
        redraw_order=False,
    ) -> Iterator[DXFGraphic]:
        """
        Yields the transformed referenced block content as virtual entities.

        This method is meant to examine the block reference entities at the target
        location without exploding the block reference.
        These entities are not stored in the entity database, have no handle and
        are not assigned to any layout. It is possible to convert these entities
        into regular drawing entities by adding the entities to the entities
        database and a layout of the same DXF document as the block reference::

            doc.entitydb.add(entity)
            msp = doc.modelspace()
            msp.add_entity(entity)

        .. warning::

            **Non-uniform scale factors** may return incorrect results for some entities
            (TEXT, MTEXT, ATTRIB).

        This method does not resolve the MINSERT attributes, only the
        sub-entities of the first INSERT will be returned. To resolve MINSERT
        entities check if multi insert processing is required, that's the case
        if the property :attr:`Insert.mcount` > 1, use the :meth:`Insert.multi_insert`
        method to resolve the MINSERT entity into multiple INSERT entities.

        This method does not apply the clipping path created by the XCLIP command.
        The method returns all entities and ignores the clipping path polygon and no
        entity is clipped.

        The `skipped_entity_callback()` will be called for all entities which are not
        processed, signature:
        :code:`skipped_entity_callback(entity: DXFEntity, reason: str)`,
        `entity` is the original (untransformed) DXF entity of the block definition, the
        `reason` string is an explanation why the entity was skipped.

        Args:
            skipped_entity_callback: called whenever the transformation of an
                entity is not supported and so was skipped
            redraw_order: yield entities in ascending redraw order if ``True``

        """
        for e in virtual_block_reference_entities(
            self,
            skipped_entity_callback=skipped_entity_callback,
            redraw_order=redraw_order,
        ):
            e.set_source_block_reference(self)
            yield e

    @property
    def mcount(self) -> int:
        """Returns the multi-insert count, MINSERT (multi-insert) processing
        is required if :attr:`mcount` > 1.

        """
        return (self.dxf.row_count if self.dxf.row_spacing else 1) * (
            self.dxf.column_count if self.dxf.column_spacing else 1
        )

    def multi_insert(self) -> Iterator[Insert]:
        """Yields a virtual INSERT entity for each grid element of a MINSERT
        entity (multi-insert).
        """

        def transform_attached_attrib_entities(insert, offset):
            for attrib in insert.attribs:
                attrib.dxf.insert += offset

        def adjust_dxf_attribs(insert, offset):
            dxf = insert.dxf
            dxf.insert += offset
            dxf.discard("row_count")
            dxf.discard("column_count")
            dxf.discard("row_spacing")
            dxf.discard("column_spacing")

        done = set()
        row_spacing = self.dxf.row_spacing
        col_spacing = self.dxf.column_spacing
        rotation = self.dxf.rotation
        for row in range(self.dxf.row_count):
            for col in range(self.dxf.column_count):
                # All transformations in OCS:
                offset = Vec3(col * col_spacing, row * row_spacing)
                # If any spacing is 0, yield only unique locations:
                if offset not in done:
                    done.add(offset)
                    if rotation:  # Apply rotation to the grid.
                        offset = offset.rotate_deg(rotation)
                    # Do not apply scaling to the grid!
                    insert = self.copy()
                    adjust_dxf_attribs(insert, offset)
                    transform_attached_attrib_entities(insert, offset)
                    yield insert

    def add_auto_attribs(self, values: dict[str, str]) -> Insert:
        """
        Attach for each :class:`~ezdxf.entities.Attdef` entity, defined in the
        block definition, automatically an :class:`Attrib` entity to the block
        reference and set ``tag/value`` DXF attributes of the ATTRIB entities
        by the ``key/value`` pairs (both as strings) of the `values` dict.
        The ATTRIB entities are placed relative to the insert location of the
        block reference, which is identical to the block base point.

        This method avoids the wrapper block of the
        :meth:`~ezdxf.layouts.BaseLayout.add_auto_blockref` method, but the
        visual results may not match the results of CAD applications, especially
        for non-uniform scaling. If the visual result is very important to you,
        use the :meth:`add_auto_blockref` method.

        Args:
            values: :class:`~ezdxf.entities.Attrib` tag values as ``tag/value``
                pairs

        """
        def unpack(dxfattribs) -> tuple[str, str, UVec]:
            tag = dxfattribs.pop("tag")
            text = values.get(tag, None)
            if text is None:  # get default value from ATTDEF
                text = dxfattribs.get("text", "")
            location = dxfattribs.pop("insert")
            return tag, text, location

        def autofill() -> None:
            for attdef in block_layout.attdefs():  # type: ignore
                dxfattribs = attdef.dxfattribs(drop={"prompt", "handle"})

                # Caution! Some mandatory values may not exist!
                # These are DXF structure errors, but loaded DXF files may have errors!
                if "tag" not in dxfattribs:
                    # ATTRIB without "tag" makes no sense!
                    logger.warning(
                        f"Skipping {str(attdef)}: missing mandatory 'tag' attribute"
                    )
                    continue
                if "insert" not in dxfattribs:
                    # Don't know where to place the ATTRIB entity.
                    logger.warning(
                        f"Skipping {str(attdef)}: missing mandatory 'insert' attribute"
                    )
                    continue

                tag, text, location = unpack(dxfattribs)
                attrib = self.add_attrib(tag, text, location, dxfattribs)
                if attdef.has_embedded_mtext_entity:
                    mtext = attdef.virtual_mtext_entity()
                    mtext.text = text
                    attrib.embed_mtext(mtext)
                attrib.transform(m)

        block_layout = self.block()
        if block_layout is not None:
            m = self.matrix44()
            autofill()
        return self

    def audit(self, auditor: Auditor) -> None:
        """Validity check."""
        super().audit(auditor)
        doc = auditor.doc
        if doc and doc.blocks:
            name = self.dxf.name
            if name is None:
                auditor.fixed_error(
                    code=AuditError.UNDEFINED_BLOCK_NAME,
                    message=f"Deleted entity {str(self)} without a BLOCK name",
                )
                auditor.trash(self)
            elif name not in doc.blocks:
                auditor.fixed_error(
                    code=AuditError.UNDEFINED_BLOCK,
                    message=f"Deleted entity {str(self)} without required BLOCK"
                    f" definition.",
                )
                auditor.trash(self)

    def __referenced_blocks__(self) -> Iterable[str]:
        """Support for the "ReferencedBlocks" protocol."""
        block = self.block()
        if block is not None:
            return (block.block_record_handle,)
        return tuple()



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
# created 2019-02-15
import pytest

from ezdxf.entities.lwpolyline import LWPolyline
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.lldxf.const import DXFAttributeError, DXFStructureError

LWPOLYLINE = """0
LWPOLYLINE
5
0
330
0
100
AcDbEntity
8
0
100
AcDbPolyline
90
0
70
0
43
0.0
"""

LWPOLYLINE_ERR = """0
LWPOLYLINE
5
0
330
0
100
AcDbEntity
8
0
"""


@pytest.fixture
def entity():
    return LWPolyline.from_text(LWPOLYLINE)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "LWPOLYLINE" in ENTITY_CLASSES


def test_default_init():
    dxfclass = LWPolyline()
    assert dxfclass.dxftype() == "LWPOLYLINE"
    assert dxfclass.dxf.handle is None
    assert dxfclass.dxf.owner is None
    assert dxfclass.dxf.hasattr("const_width") is False


def test_default_new():
    entity = LWPolyline.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": 7,
            "const_width": 2,
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.const_width == 2.0
    assert len(entity) == 0
    entity.append((1, 2))
    assert len(entity) == 1
    assert entity.dxf.count == 1
    with pytest.raises(DXFAttributeError):
        # count not writeable
        entity.dxf.count = 2


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.count == 0
    assert entity.dxf.const_width == 0


def test_write_dxf():
    entity = LWPolyline.from_text(LWPOLYLINE)
    collector = TagCollector()
    entity.export_dxf(collector)
    assert len(collector.tags) == 0, "do not export empty polylines"
    entity.append((1, 2))
    entity.export_dxf(collector)
    result = collector.tags
    expected = basic_tags_from_text(RESULT1)
    assert result == expected


def test_has_const_width():
    pline = LWPolyline()
    assert pline.has_width is False
    pline.dxf.const_width = 1
    assert pline.has_width is True


def test_has_any_start_width():
    pline = LWPolyline()
    assert pline.has_width is False
    pline.lwpoints.append((0, 0, 0.1, 0, 0))
    assert pline.has_width is True


def test_has_any_end_width():
    pline = LWPolyline()
    assert pline.has_width is False
    pline.lwpoints.append((0, 0, 0, 0.1, 0))
    assert pline.has_width is True


def test_has_arc():
    pline = LWPolyline()
    assert pline.has_arc is False
    pline.lwpoints.append((0, 0, 0, 0, 1))
    assert pline.has_arc is True


def test_raises_dxf_structure_error_for_missing_AcDbPolyline_subclass():
    with pytest.raises(DXFStructureError):
        LWPolyline.from_text(LWPOLYLINE_ERR)


RESULT1 = """0
LWPOLYLINE
5
0
330
0
100
AcDbEntity
8
0
100
AcDbPolyline
90
1
70
0
43
0.0
10
1.
20
2.
"""
